{{
PropXMM-SFE-Board.spin
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below
}}

{{
A 1 Propeller 2 COG SRAM random access demo.
Propeller P12..31 are free for user.
-----------------------------------------------------------------------------------------
                                                                                       
Design assumes clocked and latched address interface for 1MB SRAM
D0..7 on P0..7
Should use 220 Ohm+ series resistors on SRAM D0..7
A2..11 on P0..9 clocked on 74lvt574 and 74lvt278 (hi bit serves as chip select).
A12..19 on P0..7 latched on 74lvt573
A0..1 on P8..9
OE on P10 - required to keep address from corruption during latch cycle for reads
ALE on P11 - clock/latch and stepper cog signal
WE on P28 - write enable controlled by dira and ctra registers. 
}}

'{{

con
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  PAUSE    = 100
  
obj
  's : "FullDuplexSerial"
  s : "Tv_Text"

pub main | n

  's.start(31,30,0,115200)
  s.start(20)
  waitcnt(clkfreq*2+cnt)
  
  s.str(string($d,"PropXMM-SFE Board Demo",$d))
  address := $fffff
  
  stepper := @cog2
  cognew(@cog1, @command)
  waitcnt(clkfreq/10+cnt)
    
{ debug code ... leave commented for testing
  cognew(@testcog, @command)
  repeat
'}

{ debug code ... leave commented for testing
  address := 0
  'n :=  $88442200
  n~
  repeat
    n++
    data := $88442211 '+ n
    command := CMD_WR32
    repeat while command
    data~
    command := CMD_RD32
    repeat while command
    s.hex(data, 8)
    s.out(" ")
    if n // 4 == 3
      s.out($d)
    if n // 48 == 47
      s.out(0)
    'address++
'}    


{ debug code ... leave commented for testing
  'pattern test   
  repeat
    'waitcnt(clkfreq/PAUSE+cnt)
    writelong(0, $33cc6699)
    result := readlong(0)
    s.hex(result, 8)
    s.out(" ")
'{
    'waitcnt(clkfreq/PAUSE+cnt)
    writelong(0, $12345678)
    result := readlong(0)
    s.hex(result, 8)
    s.out(" ")

    s.out(0)
    s.out($d)
'}      
'}

'{ incremental test
  repeat
    s.str(string($d,"writing ... "))      
    repeat address from 0 to constant(ADDRANGE/ADDRDIV) step 4 
      data := address
      command := CMD_WR32       ' fast way

{ debug code ... leave commented for testing
      writelong(address, address)
      data := readlong(address)
      if data <> address
        s.str(string($d,"write error "))
        s.hex(address, 8)
        s.out(" ")
        s.hex(data, 8)
        s.out($d)
        repeat
          command := CMD_WR32
          command := CMD_RD32
'}

      ifnot address // constant(ADDRANGE/ADDRDIV/64)
        s.out(".")

    s.out(" ")
    s.hex(address, 8)
    
    s.str(string($d,"reading ... "))      
    repeat address from 0 to constant(ADDRANGE/ADDRDIV) step 4
      'waitcnt(clkfreq/PAUSE+cnt)
      'result := readlong(address)
      command := CMD_RD32
      repeat while command
      'result := data
'{      
      if data <> address
        s.hex(address, 8)
        s.out(" ")
        s.hex(data, 8)
        s.out($d)
'{        
        repeat  ' if error repeat on failed address
          data := address
          command := CMD_WR32
          'repeat while command
          command := CMD_RD32
          'repeat while command
'}        
      ifnot address // constant(ADDRANGE/ADDRDIV/64)
        s.out(".")

    s.out(" ")
    s.hex(address, 8)
'}

con
  ADDRDIV = $1
'}}
  
con
  DATAMASK      = $000000ff
  ADDRANGE      = $000fffff        ' demultiplexed range 14+6 = 20 bits 
  ADDRMASK      = $000003ff        ' rising ALE clocks 10 bits. falling ALE latches 8 bits. 2 bits direct connect. 
  ADR2MASK      = $00000300        ' rising ALE latches 14 bits. 6 bits address direct connect. 
  A0PIN         = 8
  A1PIN         = 9
  WEPIN         = 28
  SOPIN         = 30
  CS_ALEPIN     = 10
  OEPIN         = 11

  CMD_RD32      = 1
  CMD_WR32      = 2

var
  long  command
  long  stepper
  long  address
  long  data   
  
pub writelong(maddr, mdat)
  data := mdat
  address := maddr
  command := CMD_WR32
  repeat while command

pub readlong(maddr)
  address := maddr
  command := CMD_RD32
  repeat while command
  return data
  
dat
cog1           org       0

init

  mov          sp,      par
  add          sp,      #4
  mov          ap,      par
  add          ap,      #8
  mov          dp,      par
  add          dp,      #12

  ' start the second cog
  rdlong       val,     sp   ' get stepper cog pointer
  shr          val,     #2   ' zero 2 LSB
  shl          val,     #4   ' get function address into the right place
  or           val,     #8   ' get a new COG
  coginit      val      wr,wc,wz

  mov          val,     adrmask
  shl          val,     #8
  add          val,     cnt
  waitcnt      val,     val nr
   
  or           dira,    adatmask
  or           dira,    adrmask
  or           dira,    somask
  or           dira,    cs_ale 
  andn         dira,    wren    ' WEN has pullup ... set to output for write
  andn         dira,    outen   ' disassert OE ... assert only for SRAM read

  wrlong       ctrb,    sp      ' clear signal

  mov          frqa,    frqav
  mov          phsa,    phsav   ' 6 make sure phsa is set for each access
  mov          ctra,    ctrav   ' 7 enable counters for write strobe

  or           dira,    TESTPIN
  andn         outa,    TESTPIN
  
loop
  rdlong       cmd,     par wz
  if_z  jmp    #loop
  rdlong       addr,    ap
  test         cmd,     #CMD_RD32 wz
  if_nz jmp    #rd32
  test         cmd,     #CMD_WR32 wz
  if_nz rdlong val,     dp
  if_nz jmp    #wr32

done
  wrlong       dirb,    par     ' ctrb should be zero    
  jmp          #loop

'------------------------------------------------------------------------
' rd32 
' incoming address must be long aligned (bits A[0:1] = 0)
rd32
  or           dira,    adatmask' make sure all addr/data bits are outputs
  shr          addr,    #2      ' 0000.0000.0000.00ji.hgfe.dcba.9876.5432
  mov          outa,    addr    ' 
  andn         outa,    cs_ale  ' clear ale for clocking in bits 2..11
  or           outa,    cs_ale  ' clock the bits
  andn         dira,    cs_ale  ' clock the bits
  shr          addr,    #10     ' 0000.0000.0000.0000.0000.0000.jihg.fedc
  mov          outa,    addr    ' 
  or           dira,    cs_ale  ' CS and ALE bits are 0 by now
  or           dira,    oe_ale  ' OE and ALE bits are 0 by now
  andn         dira,    datmask ' data for read
  nop
  movs         val,     ina     ' xxxxxxxxxxxxxxxxxxxxxxxaAAAAAAAA .
  ror          val,     #17 wc  ' xxxxxxxxaAAAAAAAAxxxxxxxxxxxxxxx A0
  movi         val,     ina     ' bBBBBBBBBAAAAAAAAxxxxxxxxxxxxxxx A0
  shr          val,     #8      ' ........bBBBBBBBBAAAAAAAAxxxxxxx A0
  movi         val,     ina     ' cCCCCCCCCBBBBBBBBAAAAAAAAxxxxxxx A0
  shr          val,     #8      ' ........cCCCCCCCCBBBBBBBBAAAAAAA A0
  movi         val,     ina     ' dDDDDDDDDCCCCCCCCBBBBBBBBAAAAAAA A0
  rcl          val,     #1      ' DDDDDDDDCCCCCCCCBBBBBBBBAAAAAAAA A0
  andn         dira,    outen   ' disassert OE
  wrlong       val,     dp      ' normally for LMM mode this is not necessary since kernel runs in COG
  jmp   #done  'ret             ' 23 ticks required for LMM: 1150ns or 3.5MB/s                         

'------------------------------------------------------------------------
' wr32
' We clock in the phase address bits.
' Addr cog latches second phase address bits.
wr32
  or           dira,    adatmask' all address/data to outputs 
  shr          addr,    #2      ' 0000.0000.0000.00ji.hgfe.dcba.9876.5432
  mov          outa,    addr    ' 
  andn         outa,    cs_ale  ' clear ale for clocking in bits 2..11
  or           outa,    cs_ale  ' clock the bits
  andn         dira,    cs_ale  ' clock the bits
  shr          addr,    #10     ' 0000.0000.0000.0000.0000.0000.jihg.fedc
  mov          outa,    addr    ' 
  or           dira,    cs_ale  ' CS and ALE bits are 0 by now
  andn         dira,    x300
  movs         outa,    val
  or           dira,    wren    '
  shr          val,     #8      ' next byte
  movs         outa,    val
  shr          val,     #8      ' next byte
  movs         outa,    val
  shr          val,     #8      ' next byte
  movs         outa,    val
  'nop
  or           outa,    wren    ' sharp edge
  andn         dira,    wren    '
  jmp          #done    'ret    ' 21 ticks 1050ns 3.8MB/s                               

TESTPIN        long     $8000000

sigdat         long     $4
adrmask        long     ADDRMASK
adrmask2       long     ADR2MASK
datmask        long     DATAMASK
adatmask       long     DATAMASK | $300
oedat          long     DATAMASK | 1 << OEPIN
wedat          long     DATAMASK | 1 << WEPIN
x200           long     $200
x300           long     $300
we_ale         long     1 << CS_ALEPIN | 1 << WEPIN
oe_ale         long     1 << CS_ALEPIN | 1 << OEPIN
cs_ale         long     1 << CS_ALEPIN
caledat        long     DATAMASK | 1 << CS_ALEPIN
wren           long     1 << WEPIN
outen          long     1 << OEPIN
somask         long     1 << SOPIN

ctrav          long     %00100_000 << 23 + WEPIN
frqav          long     $20000000
phsav          long     0

val            res      1
val2           res      1
addr           res      1
cmd            res      1
tp             res      1
sp             res      1
ap             res      1
dp             res      1


dat
cog2           org      0
  or           dira,    xx300
  mov          outa,    #0      ' make sure bits are clear
  
adr2loop
  waitpne      cale,    cale
  waitpeq      cale,    cale
  andn         outa,    #0      ' make sure all bits are clear so data cog can work
  nop                           ' wait for main routine
  nop                           ' wait for main routine
  nop                           ' wait for main routine
  nop                           ' wait for main routine
  nop                           ' wait for main routine
  add          outa,    #$100   ' ok, step address
  nop                           ' wait
  add          outa,    #$100   ' ok, step address
  nop                           ' wait
  add          outa,    #$100   ' ok, step address
  nop                           ' wait
  nop                           ' wait
  mov          outa,    #0      ' make sure bits are clear next address is good
  jmp          #adr2loop        ' 600ns                                                

cale           long     1 << CS_ALEPIN
oen2           long     1 << OEPIN
dmask          long     $ff
xx300          long     $300 
junk           res      1
ap2            res      1
sp2            res      1

               
'{ fast test code for debug
DAT
testcog        org 0

  mov          msp,     par
  add          msp,     #4
  mov          map,     msp
  add          map,     #4
  mov          mdp,     map
  add          mdp,     #4

testloop
'{  
  mov          testadr, adrstart
  'mov          testpat, #0
  mov          testcnt, testlen  

testwrite
  wrlong       testpat, mdp      
  wrlong       testadr, map
  mov          mval,    #CMD_WR32
  wrlong       mval,    par
testwait1
  rdlong       mval,    par wz
  if_nz jmp    #testwait1
  
  'add          testpat, #1
  'add          testadr, #4
  'and          testadr, mbmask

  djnz         testcnt, #testwrite
'}  
'{

  mov          testadr, adrstart
  mov          testcnt, testlen
    
testread
  wrlong       testadr, map
  mov          mval,    #CMD_RD32
  wrlong       mval,    par
testwait2
  rdlong       mval,    par wz
  if_nz jmp    #testwait2
  'add          testadr, #4
  'and          testadr, mbmask
  
  djnz         testcnt, #testread
'}  
  jmp          #testloop
  
testcnt        long 1
testlen        long 1
mbmask         long $fffff
testpat        long $88442211
testadr        long 0
adrstart       long $4

msp     res 1
map     res 1
mdp     res 1
mtp     res 1
mcmd    res 1
mval    res 1
mda     res 1
'}

{
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                  
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
}